/* $Id: globus_replica_catalog_sets.c,v 1.24 2001/12/21 02:42:28 meder Exp $ */

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
/**
 * @file
 *
 * Globus Replica Catalog Sets C API Source File
 *
 * This file defines functions that manage sets of entries and
 * attributes.
 *
 * @see globus_replica_catalog_sets.h
 */
#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */

#include "globus_replica_catalog_sets.h"

/* Attribute Set Management */

globus_result_t
globus_replica_catalog_attribute_set_init(
	globus_replica_catalog_attribute_set_t * attributes )
{
    attributes->collection = NULL;
    attributes->entry = NULL;
    attributes->result = NULL;
    attributes->ber = NULL;
    attributes->attribute_name = NULL;

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_attribute_set_destroy(
	globus_replica_catalog_attribute_set_t * attributes )
{
    attributes->collection = NULL;
    attributes->entry = NULL;
    if (attributes->result != NULL) {
	ldap_msgfree(attributes->result);
	attributes->result = NULL;
    }
    if (attributes->ber != NULL) {
	ber_free(attributes->ber, 0);
	attributes->ber = NULL;
    }
    if(attributes->attribute_name != NULL)
    {
        ber_memfree(attributes->attribute_name);
        attributes->attribute_name = NULL;
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_attribute_set_first(
	globus_replica_catalog_attribute_set_t * attributes )
{
    int lderr;

    if (attributes->entry == NULL) {
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"No entry" );
    }

    if ( ( attributes->attribute_name = ldap_first_attribute(
	    attributes->collection->ld,
	    attributes->entry,
	    &(attributes->ber) ) ) == NULL ) {
	ldap_get_option(attributes->collection->ld, LDAP_OPT_ERROR_NUMBER,
		&lderr);
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_attribute_set_next(
	globus_replica_catalog_attribute_set_t * attributes )
{
    int lderr;

    if (attributes->ber == NULL) {
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"No attributes" );
    }

    if ( ( attributes->attribute_name = ldap_next_attribute(
	    attributes->collection->ld,
	    attributes->entry,
	    attributes->ber ) ) == NULL ) {
	ldap_get_option(attributes->collection->ld, LDAP_OPT_ERROR_NUMBER,
		&lderr);

        if(lderr == LDAP_SUCCESS)
        {
            return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"No more attributes");
        }
        else
        {
            return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
        }
    }

    /* skip filenames */

    /* don't really know why we want to skip filenames,
     * this causes problems - Sam
       
    while ( (attributes->attribute_name != NULL)
	    && ( strcmp( attributes->attribute_name,
		    GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE )
			    == 0 ) ) {
	if ( ( attributes->attribute_name = ldap_next_attribute(
		attributes->collection->ld,
		attributes->entry,
		attributes->ber ) ) == NULL ) {
	    ldap_get_option(attributes->collection->ld, LDAP_OPT_ERROR_NUMBER,
		    &lderr);

            if(lderr == LDAP_SUCCESS)
            {
                return GLOBUS_REPLICA_CATALOG_ERROR(
                    GLOBUS_NULL,
                    "No more attributes");
            }
            else
            {
                return GLOBUS_REPLICA_CATALOG_ERROR(
                    GLOBUS_NULL,
                    ldap_err2string(lderr) );
            }
        }
    }
    */
    return GLOBUS_SUCCESS;
}

globus_bool_t
globus_replica_catalog_attribute_set_more(
	globus_replica_catalog_attribute_set_t * attributes )
{
    return (attributes->ber != NULL)
	    && (attributes->attribute_name != NULL);
}

globus_result_t
globus_replica_catalog_attribute_set_get_name(
	globus_replica_catalog_attribute_set_t * attributes,
	char ** attribute_name )
{
    if (attributes->attribute_name == NULL) {
	*attribute_name = GLOBUS_NULL;
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"No attributes" );
    }

    *attribute_name = globus_libc_strdup(attributes->attribute_name);

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_attribute_set_get_values(
	globus_replica_catalog_attribute_set_t * attributes,
	char *** attribute_values )
{
    int lderr;

    if (attributes->attribute_name == NULL) {
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"No attributes" );
    }

    if ( ( *attribute_values = ldap_get_values(
	    attributes->collection->ld,
	    attributes->entry,
	    attributes->attribute_name ) ) == NULL ) {
	*attribute_values = GLOBUS_NULL;
	ldap_get_option(attributes->collection->ld, LDAP_OPT_ERROR_NUMBER,
		&lderr);
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_attribute_set_free_values(
	char ** attribute_values )
{
    ldap_value_free(attribute_values);

    return GLOBUS_SUCCESS;
}

/* Entry Set Management */

globus_result_t
globus_replica_catalog_entry_set_init(
	globus_replica_catalog_entry_set_t * entries )
{
    entries->collection = NULL;
    entries->collection_close = GLOBUS_FALSE;
    entries->result = NULL;
    entries->entry = NULL;

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_entry_set_destroy(
	globus_replica_catalog_entry_set_t * entries )
{
    if ( (entries->collection != GLOBUS_NULL)
	    && (entries->collection_close == GLOBUS_TRUE) ) {
	globus_libc_free(entries->collection->attrs);
	entries->collection->attrs = GLOBUS_NULL;
	globus_replica_catalog_collection_close(entries->collection);
	globus_libc_free(entries->collection);
	entries->collection_close = GLOBUS_FALSE;
    }
    if (entries->result != NULL) {
	ldap_msgfree(entries->result);
	entries->result = NULL;
    }
    entries->entry = NULL;

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_entry_set_first(
	globus_replica_catalog_entry_set_t * entries )
{
    int lderr;

    if (entries->result == NULL) {
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"No result" );
    }

    if ( ( entries->entry = ldap_first_entry(
	    entries->collection->ld,
	    entries->result ) ) == NULL ) {
	ldap_get_option(entries->collection->ld, LDAP_OPT_ERROR_NUMBER,
		&lderr);
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_entry_set_next(
	globus_replica_catalog_entry_set_t * entries )
{
    int lderr;

    if (entries->entry == NULL) {
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"No entry" );
    }

    if ( ( entries->entry = ldap_next_entry(
	    entries->collection->ld,
	    entries->entry ) ) == NULL ) {
	ldap_get_option(entries->collection->ld, LDAP_OPT_ERROR_NUMBER,
		&lderr);
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_bool_t
globus_replica_catalog_entry_set_more(
	globus_replica_catalog_entry_set_t * entries )
{
    return (entries->result != NULL) && (entries->entry != NULL);
}

globus_result_t
globus_replica_catalog_entry_set_count(
	globus_replica_catalog_entry_set_t * entries,
	int * count )
{
    int lderr;

    if (entries->result == NULL) {
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"No result" );
    }

    if ( ( *count = ldap_count_entries(
	    entries->collection->ld,
	    entries->result ) ) == -1 ) {
	ldap_get_option(entries->collection->ld, LDAP_OPT_ERROR_NUMBER,
		&lderr);
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_entry_set_get_dn(
	globus_replica_catalog_entry_set_t * entries,
	char ** entry_dn )
{
    int lderr;

    if (entries->entry == NULL) {
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"No entry" );
    }

    if ( (*entry_dn = ldap_get_dn( entries->collection->ld,
	    entries->entry )) != NULL ) {
	return GLOBUS_SUCCESS;
    }

    *entry_dn = GLOBUS_NULL;

    ldap_get_option(entries->collection->ld, LDAP_OPT_ERROR_NUMBER,
	    &lderr);
    if (lderr != LDAP_SUCCESS) {
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_entry_set_get_url(
	globus_replica_catalog_entry_set_t * entries,
	char ** entry_url )
{
    char * entry_dn;
    int lderr;

    if (entries->entry == NULL) {
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"No entry" );
    }

    if ( (entry_dn = ldap_get_dn( entries->collection->ld,
	    entries->entry )) == NULL ) {
	ldap_get_option(entries->collection->ld, LDAP_OPT_ERROR_NUMBER,
		&lderr);
	if (lderr != LDAP_SUCCESS) {
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		    GLOBUS_NULL,
		    ldap_err2string(lderr) );
	} else {
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		    GLOBUS_NULL,
		    "No DN" );
	}
    }

    *entry_url = (char *) globus_libc_malloc(
	    strlen(entries->collection->urld->lud_host)
	    + 5 + strlen(entry_dn) + 10 );

    sprintf( *entry_url, "ldap://%s:%d/%s",
	    entries->collection->urld->lud_host,
	    entries->collection->urld->lud_port,
	    entry_dn );

    free(entry_dn);

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_entry_set_free_dn(
	char * entry_dn )
{
    if (entry_dn != GLOBUS_NULL) {
	free(entry_dn);
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_entry_set_get_name(
	globus_replica_catalog_entry_set_t * entries,
	char ** name )
{
    char * entry_dn;
    char ** entry_rdns;
    int lderr;

    if (entries->entry == NULL) {
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"No entry" );
    }

    if ( (entry_dn = ldap_get_dn( entries->collection->ld,
	    entries->entry )) != NULL ) {
	if ((entry_rdns = ldap_explode_dn(entry_dn, 1)) != NULL) {
	    *name = globus_libc_strdup(entry_rdns[0]);
	    free(entry_dn);
	    ldap_value_free(entry_rdns);
	    return GLOBUS_SUCCESS;
	}
    }

    *name = GLOBUS_NULL;

    ldap_get_option(entries->collection->ld, LDAP_OPT_ERROR_NUMBER,
	    &lderr);
    if ( (entry_dn == NULL) && (lderr != LDAP_SUCCESS) ) {
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
    }

    free(entry_dn);

    if ( (entry_rdns == NULL)
	    && (errno != LDAP_SUCCESS) ) {
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(errno) );
    }

    ldap_value_free(entry_rdns);

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_entry_set_get_attributes(
	globus_replica_catalog_entry_set_t * entries,
	globus_replica_catalog_attribute_set_t * attributes )
{
    if (entries->entry == NULL) {
	return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"No entry" );
    }

    attributes->collection = entries->collection;
    attributes->entry = entries->entry;

    return GLOBUS_SUCCESS;
}
